#
# vocalips044.tcl - GUI
#
# A GUI wrapper written in Tcl/Tk. 
# It wraps the command-line processor (actually a Perl script
# "vsqtxt2lipsync.pl") that does the processing. 
# 
# 1) The user specifies the VSQ file name, or drag-and-drop a VSQ file.
# 2) The wrapper calls mf2t, and then vsqtxt2lipsync,
#    to display the track names and the BPM map. 
# 3) The user specifies the output mode, and the tracks to be
#    processed. Note that more than one tracks can be processed.
# 4) The wrapper calls vsqtxt2lipsync for the second time,
#    to obtain the results.
# 5) In "MusicXML" output mode, 'unit for time' is ignored;
#    tick is used. 
# 
# That's totally fine to run vsqtxt2lipsync manually from inside a
# console. All the options are described at the beginning 
# of the script. 
#
# preview window is almost working, but japanese chars are
# not showing. very weird. this is put off.
#
# Rev.0.44  Apr.  1, 2010  - MusicXML output.  A bit inconsistent idea,
#                            but just for handiness. 
#                          - "unit for time" moved to bpm-map area.
#                          - The option area is now dynamically updated.
#                          - Aprilfool fun: more color variations!
#
# Rev.0.43  Jan. 23, 2010  - Has got a radiobutton for time index mode.
#                            Slightly, it now terminates properly when 
#                            the right-top "x" button is clicked.
#                            Also slightly, the newest phoneme-conversion
#                            table is chosen. 
# Rev.0.42  Apr. 01, 2009  - For lyrics other than Japanese, 
#                            "standard, raw-phonetics" mode is
#                            introduced.
#                          - drag-and-drop is now supported for VSQ
#                            files.
#                          - conversion table-name is always shown
#                            as "T.TBX" in the output file.
#                          - Aprilfool version with more charactor
#                            color variations!
# Rev.0.41  Apr. 01, 2008  Aprilfool version: Miku/Rin/Len color
#                          pallette:-)
# Rev.0.39  Mar. 09, 2008  add beat track. 
# Rev.0.30  Dec. 24, 2007  add BPM-map preview.
# Rev.0.21  Nov. 27, 2007  just for the update to vsqtxt2lipsync_042.
# Rev.0.20  Nov. 23, 2007
# Rev.0.10  Nov. 3, 2007
# 
# author: zhuo   www.geocities.jp/zhuoware
# 

package require tkdnd


set VocalipsRevision "0.44"
set Vsqtxt2lipsyncScript "bin/vsqtxt2lipsync_052.pl"
set Vocalips2musicxmlScript "bin/vocalips2musicxml_04a.pl"

set w .top
catch {destroy $w}
toplevel $w
wm title $w "Vocalips Rev.$VocalipsRevision"
wm iconname $w "Vocalips Rev.$VocalipsRevision"
wm withdraw .

set vocalipsAttr(fixedFont) {FixedSys 8}


proc createFileEntry { path text cmd } {
    frame $path
    label $path.lab -width 16 -text $text -anchor e
    entry $path.ent -width 80
    button $path.but -text "Browse ..." -command "$cmd $path"
    pack $path.lab -side left
    pack $path.ent -side left -expand yes -fill x
    pack $path.but -side left
    pack $path -fill x -padx 1c -pady 3
}

proc setEntryString { path string } {
    if {[string compare $string ""]} {
	$path delete 0 end
	$path insert 0 $string
	$path xview end
    }
}

proc update_ui { w mode } {
    if { $mode == "m" } {
	pack forget $w.settings.opt.fb
	pack $w.settings.opt.mx -side top -padx 1m -fill x -expand no
	return 
    } 
    if { $mode == "f" } {
	pack $w.settings.opt.fb -side top -padx 1m -fill x -expand no
	pack forget $w.settings.opt.mx
	return 
    } 
    pack forget $w.settings.opt.fb
    pack forget $w.settings.opt.mx
}



createFileEntry $w.vsqfile "vocaloid file: " "vsqFileChooser $w.outfile.ent"
createFileEntry $w.outfile "output file: " "outFileChooser"
createFileEntry $w.tblfile "phoneme conv. file: " "tblFileChooser"

$w.tblfile.ent delete 0 end
$w.tblfile.ent insert 0 [lindex [glob -nocomplain -types f "*.tbl"] end] 
$w.tblfile.ent xview end


frame $w.time
labelframe $w.time.time -text "unit for time" -padx 2 -pady 2
radiobutton $w.time.time.sec -text "second" -variable outputTimeInSec -value 1
radiobutton $w.time.time.tick -text "tick" -variable outputTimeInSec -value 0
$w.time.time.sec select

pack $w.time.time.sec -side left -padx 1m -expand yes
pack $w.time.time.tick -side right -padx 1m  -expand yes
pack $w.time.time -side top -fill x -expand no


labelframe $w.time.bpmmap -text "bpm map" -padx 5 -pady 5
scrollbar $w.time.bpmmap.scroll -command "$w.time.bpmmap.list yview"
text $w.time.bpmmap.list -yscroll "$w.time.bpmmap.scroll set" -setgrid 1 -width 20 -height 10 -relief flat -font FixedSys  -state disabled

pack $w.time.bpmmap.scroll -side right -fill y
pack $w.time.bpmmap.list -side left -fill y -expand yes
pack $w.time.bpmmap -side top -fill y -expand yes

pack $w.time -side left -padx 3m -fill y -expand no


frame $w.settings
labelframe $w.settings.tracks -text "track chooser" -padx 5 -pady 5
scrollbar $w.settings.tracks.scroll -command "$w.settings.tracks.list yview"
listbox $w.settings.tracks.list -yscroll "$w.settings.tracks.scroll set" -setgrid 1 -width 24 -height 8 -selectmode extended
pack $w.settings.tracks.scroll -side right -fill y
pack $w.settings.tracks.list -side left -fill both -expand yes 
pack $w.settings.tracks -side left -padx 3m -fill both -expand yes 


labelframe $w.settings.mode -text "output mode" -padx 5 -pady 5
radiobutton $w.settings.mode.e2 -text "standard" -variable outputMode -relief flat -value "e2" -anchor w -command "update_ui $w e2"
radiobutton $w.settings.mode.e3 -text "standard, raw-phonetics" -variable outputMode -relief flat -value "e3" -anchor w -command "update_ui $w e3"
radiobutton $w.settings.mode.e1 -text "event-base, \"sil\" when early" -variable outputMode -relief flat -value "e1" -anchor w -command "update_ui $w e1"
radiobutton $w.settings.mode.e0 -text "event-base, always with \"end\"" -variable outputMode -relief flat -value "e0" -anchor w -command "update_ui $w e0"
radiobutton $w.settings.mode.v -text "raw vsq text" -variable outputMode -relief flat -value "v" -anchor w -command "update_ui $w v"
radiobutton $w.settings.mode.f -text "frame-base" -variable outputMode -relief flat -value "f" -anchor w -command "update_ui $w f"
radiobutton $w.settings.mode.m -text "MusicXML" -variable outputMode -relief flat -value "m" -anchor w -command "update_ui $w m"
pack $w.settings.mode.e2 $w.settings.mode.e3 $w.settings.mode.e1 $w.settings.mode.e0 $w.settings.mode.v $w.settings.mode.f $w.settings.mode.m -side top -pady 2 -expand no -fill x
pack $w.settings.mode -side left -expand no -fill y  -padx 3m
$w.settings.mode.e2 invoke




labelframe $w.settings.msg -text "notes" -padx 5
label $w.settings.msg.note -text "Vocalips Rev.$VocalipsRevision\t\tApril 1, 2010

- You can generate MusicXML! (But be sure to
  convert the file to utf-8 encoding manually! )

- Click this area to enjoy more color variations!

- You can drag&drop a VSQ file to this window!
" -anchor nw -justify left

pack $w.settings.msg.note -side top -padx 2m -pady 3m -fill x -expand no
pack $w.settings.msg -side top -padx 3m -fill x -expand no


proc addOptionEntry { path text val row } {
    set labpath [format "%slab" $path ]
    set entpath [format "%sent" $path ]
    label $labpath -text $text
    entry $entpath -width 8 -justify right 
    $entpath insert 0 $val
    grid $labpath -row $row -column 0 -sticky e -padx 5 -pady 5
    grid $entpath -row $row -column 1 -sticky w -padx 5 -pady 5
}

labelframe $w.settings.opt -text "output mode options" -padx 5

frame $w.settings.opt.fb
addOptionEntry $w.settings.opt.fb.fps "fps" "30.00" 0
addOptionEntry $w.settings.opt.fb.offset "offset(msec)" "0.00" 1

frame $w.settings.opt.mx

addOptionEntry $w.settings.opt.mx.bpm  "bpm" "120" 0
addOptionEntry $w.settings.opt.mx.numer "time sig up" "4" 1
addOptionEntry $w.settings.opt.mx.denom "time sig dn" "4" 2
addOptionEntry $w.settings.opt.mx.qnt   "quantize" "120" 3
addOptionEntry $w.settings.opt.mx.gap   "connect" "960" 4
addOptionEntry $w.settings.opt.mx.pshift "pitch shift" "0" 5

#pack is done when the radiobuttons are pushed
#pack $w.settings.opt.fb -side top -padx 1m -fill x -expand no
#pack $w.settings.opt.mx -side top -padx 1m -fill x -expand no

pack $w.settings.opt -side top -padx 3m -fill both -expand yes
pack $w.settings -side top -fill both -expand yes 


frame $w.buttons
#checkbutton $w.buttons.doPreview -text "preview"
button $w.buttons.convert -width 16 -text "convert" -command "convert $w"
button $w.buttons.quit -width 16 -text quit -command "destroy $w; exit"
#pack $w.buttons.doPreview -side left -expand 1
pack $w.buttons.convert -side left -expand 1
pack $w.buttons.quit -side left -expand 1
pack $w.buttons -side bottom -fill x -pady 5m

wm protocol $w WM_DELETE_WINDOW "destroy $w; exit"

dnd bindtarget .top text/uri-list <Drop> {
    global w
    vsqFilePreload [string trim "%D"] $w.outfile.ent $w.vsqfile
    cleanWorkFiles
}







set colorModeIndex -1
set colorModes { None Miku RinLen Luka Kaito Meiko Gackpo Gumi }
array set colorValues {
    Miku,MonoDark \#373b3e
    Miku,MonoMid \#bec8d1
    Miku,MonoLight \#fffeec
    Miku,Col1Light \#c3e5e7
    Miku,Col1Mid \#86cecb
    Miku,Col1Dark \#137a7f
    Miku,Col2 \#e12885
    RinLen,MonoDark \#555658
    RinLen,MonoMid \#b7b7b7
    RinLen,MonoLight \#ffffff
    RinLen,Col1Light \#fffbe2
    RinLen,Col1Mid \#fcf5a7
    RinLen,Col1Dark \#ffd98c
    RinLen,Col2 \#2b90a2
    Luka,MonoDark \#35393c
    Luka,MonoMid \#4b342e
    Luka,MonoLight \#d8ca8b
    Luka,Col1Light \#fffcfb
    Luka,Col1Mid \#ebd3cf
    Luka,Col1Dark \#c89697
    Luka,Col2 \#4ebfe9
    Kaito,MonoDark \#423839
    Kaito,MonoMid \#9a8c69
    Kaito,MonoLight \#ede0c0
    Kaito,Col1Light \#cdcfe8
    Kaito,Col1Mid \#aeb6e5
    Kaito,Col1Dark \#413a87
    Kaito,Col2 \#f29c49
    Meiko,MonoDark \#892e2d
    Meiko,MonoMid \#cb213c
    Meiko,MonoLight \#ffcd98
    Meiko,Col1Light \#977e46
    Meiko,Col1Mid \#615231
    Meiko,Col1Dark \#51322f
    Meiko,Col2 \#b3b5b4
    Gackpo,MonoDark \#4b4e77
    Gackpo,MonoMid \#00c5e8
    Gackpo,MonoLight \#bef9f1
    Gackpo,Col1Light \#ffffff
    Gackpo,Col1Mid \#ffca3d
    Gackpo,Col1Dark \#e86664
    Gackpo,Col2 \#a170e4
    Gumi,MonoDark \#ef7922
    Gumi,MonoMid \#f4f4b6
    Gumi,MonoLight \#fffeff
    Gumi,Col1Light \#46b9a8
    Gumi,Col1Mid \#359d22
    Gumi,Col1Dark \#6e7377
    Gumi,Col2 \#be1453
    None,MonoDark \#ecead9
    None,MonoMid \#ecead9
    None,MonoLight \#ffffff
    None,Col1Light \#ecead9
    None,Col1Mid \#ecead9
    None,Col1Dark \#ecead9
    None,Col2 \#ecead9
}
#color info from: http://smallwebmemo.blog113.fc2.com/blog-entry-311.html
#gumi : hATrayflood from PIAPRO: http://piapro.jp/content/j1n745hxi4xlroxl

bind $w.settings.msg.note <1> {
    setColorMode $w
}

proc setColorMode {w} {
    global colorModeIndex colorModes colorValues
    incr colorModeIndex
    if {$colorModeIndex >= [llength $colorModes]} {
	set colorModeIndex 0
    }
    set mode [lindex $colorModes $colorModeIndex]

    setColorsRecursive [winfo toplevel $w] $mode
}

proc setColorsRecursive {w mode} {
    global colorValues
    switch -exact [winfo class $w] {
	Toplevel {
	    $w configure -background $colorValues($mode,Col1Dark)
	}
	Frame -
	Labelframe {
	    $w configure -background $colorValues($mode,Col1Mid)
	}
	Label {
	    $w configure -background $colorValues($mode,Col1Light)
	}
	Button -
	Radiobutton {
	    $w configure -background $colorValues($mode,Col1Light) \
		-activebackground $colorValues($mode,Col2)
	}
	Entry -
	Listbox -
	Text {
	    $w configure -background $colorValues($mode,MonoLight)
	}
	Scrollbar {
	    $w configure -background $colorValues($mode,MonoMid)
	}
    }
    foreach s [winfo children $w] {
	setColorsRecursive $s $mode
    }
}

setColorMode $w


proc getSelectedTracks {} {
    global w
    return [$w.settings.tracks.list curselection]
}


## the return value is a list like:
##     { A B }
## where  A is the seleted tracks, either: 
##     {}          # no tracks specified
## or 
##     { 0 1 3 }   # these are specified
## B is 0  when the beat track is NOT specified, 1 otherwise.
##
proc getSelectedTracksInDetail {} {
    global w
    set maxIdx [expr [$w.settings.tracks.list size]-1]
    set tracks [$w.settings.tracks.list curselection]
    set realTracks [lsearch -all -inline -exact -not $tracks $maxIdx]
    set hasBeatTrack [llength [lsearch -inline -exact $tracks $maxIdx]]
    return [list $realTracks $hasBeatTrack]
}


proc setTrackList {trackListFile} {
    global w Vsqtxt2lipsyncScript
    set listpath $w.settings.tracks.list

    set fd [open "$trackListFile"]
    $listpath delete 0 end
    while {[gets $fd line] >= 0 } {
	set elms [split $line "\t"]
	$listpath insert end [lindex $elms 1]
    }
    close $fd
## new in 0.39
    $listpath insert end "*BEAT*"
## note:  japanese characters were not shown....
##        this might take place when track names are in Japanese, too.
##        we would need care.
## end
    $listpath selection set 0 end
}



proc setBPMMap {bpmFile} {
    global w Vsqtxt2lipsyncScript
    set listpath $w.time.bpmmap.list

    set isBPMSection 0
    set isTimeSigSection 0

    set fd [open "$bpmFile"]
    while {[gets $fd line] >= 0 } {
	set elms [split $line "\t"]
	if { [lindex $elms 0] == "\#\#" } {
	    switch [lindex $elms 1] {
		"begin" { set val 1 }
		"end" { set val 0 }
	    }
	    switch [lindex $elms 2] {
		"bpm_map" { set isBPMSection $val }
		"timesig_map" { set isTimeSigSection $val }
	    }
	    continue
	}

	if { $isBPMSection } { lappend BPMLines [join $elms ""] }
	if { $isTimeSigSection } { lappend TimeSigLines [format "%s   %2d / %2d" [lindex $elms 0] [lindex $elms 1] [lindex $elms 2] ] }
    }
    close $fd

    $listpath configure -state normal
    $listpath delete 0.0 end
    $listpath insert end " time            BPM\n"
    $listpath insert end "--------------------\n"
    foreach BPMLine $BPMLines {
	$listpath insert end $BPMLine
	$listpath insert end "\n"
    }
    $listpath insert end "\n\n time        TimeSig\n"
    $listpath insert end "--------------------\n"
    foreach TSLine $TimeSigLines {
	$listpath insert end $TSLine
	$listpath insert end "\n"
    }

    $listpath configure -state disabled
}


proc vsqFileChooser { out_ent path } {
    set types {{"vocaloid files" {.vsq}} {"All files" *}}
    set file "[tk_getOpenFile -filetypes $types -parent $path -initialfile [$path.ent get]]"
    vsqFilePreload $file $out_ent $path
}

proc vsqFilePreload { file out_ent path } {
    vsqFilePreloadInternal $file $out_ent $path 
    cleanWorkFiles
}

proc vsqFilePreloadInternal { file out_ent path } {
    global w  Vsqtxt2lipsyncScript outputTimeInSec
    if { "$file" == "" || ! [file exists "$file"] } {
	return
    }

    regsub {\.vsq$} "$file" ".txt" outfile 
    setEntryString $path.ent "$file"
    setEntryString $out_ent "$outfile"

    file copy -force "$file" "T.VSQ"
    if { [catch {exec bin/mf2t "T.VSQ" "T.MFT"} msg ] } {
	tk_messageBox -icon info -message "mf2t error: $msg" -type ok -parent $w
	return
    }
    if { [catch {exec bin/Perl $Vsqtxt2lipsyncScript -m t < "T.MFT" > "T1.TXT"} msg ] } {
	tk_messageBox -icon info -message "track-info error: $msg" -type ok -parent $w
	return
    }
    setTrackList "T1.TXT"


    set cmd "exec bin/Perl $Vsqtxt2lipsyncScript "
    if { $outputTimeInSec == 0 } { append cmd "-i " }
    append cmd "-m m < T.MFT > T2.TXT "
    if { [catch {eval $cmd} msg] } {
	tk_messageBox -icon info -message "bpm-info error: $msg" -type ok -parent $w
	catch {file delete "T.VSQ" "T.MFT" "T1.TXT" "T2.TXT"}
	return
    }
    setBPMMap "T2.TXT"
}


proc tblFileChooser { path } {
    set types {{"table files" {.tbl}} {"All files" *}}
    setEntryString $path.ent "[tk_getOpenFile -filetypes $types -parent $path -initialfile [$path.ent get]]"
}

proc outFileChooser { path } {
    set types {{"text files" {.txt}} {"All files" *}}
    setEntryString $path.ent "[tk_getSaveFile -filetypes $types -parent $path -initialfile [$path.ent get]]"
}

proc convert { w } {
    convertInternal $w
    cleanWorkFiles
}


proc convertInternal { w } {
    global outputMode outputTimeInSec Vsqtxt2lipsyncScript Vocalips2musicxmlScript
# global
    set vsqfile "[$w.vsqfile.ent get]"
    set tblfile "[$w.tblfile.ent get]"
    set outfile "[$w.outfile.ent get]"

    set fps [$w.settings.opt.fb.fpsent get]
    set offset [$w.settings.opt.fb.offsetent get]

    set bpm [$w.settings.opt.mx.bpment get]
    set numer [$w.settings.opt.mx.numerent get]
    set denom [$w.settings.opt.mx.denoment get]
    set pshift [$w.settings.opt.mx.pshiftent get]
    set qnt [$w.settings.opt.mx.qntent get]
    set gap [$w.settings.opt.mx.gapent get]

    set outputModeOpt $outputMode
    set outputTimeInSecOpt $outputTimeInSec
    if { $outputMode == "m" } {
	set outputModeOpt "e3"
	set outputTimeInSecOpt 0
    }

    set trackInfo [getSelectedTracksInDetail]
    set tracks [join [lindex $trackInfo 0] ","]
    if { $outputMode == "m" } {
	set tracks [lindex [lindex $trackInfo 0] 0]
    }
    if { $tracks == "" } {
	set tracks "none"
    }

    set beatOption ""
    if { [lindex $trackInfo 1] == 1 } {
	set beatOption "-b"
    }

    if { "$vsqfile" == "" || ! [file exists "$vsqfile"] } {
	tk_messageBox -icon info -message "Please choose a vocaloid file." -type ok -parent $w
	return
    }
    if { "$tblfile" == "" || ! [file exists "$tblfile"] } {
	tk_messageBox -icon info -message "Please choose a table file." -type ok -parent $w
	return
    }
    if { "$outfile" == "" } {
	tk_messageBox -icon info -message "Please choose an output file." -type ok -parent $w
	return
    }
    if { $outputModeOpt == "" } {
	tk_messageBox -icon info -message "Please choose an output mode." -type ok -parent $w
	return
    }

    if { (![file exists "$outfile"]) || ([tk_messageBox -icon question -message "file exists: overwrite? $outfile" -type okcancel -parent $w] == "ok") } {
	file copy -force "$vsqfile" "T.VSQ"
	file copy -force "$tblfile" "T.TBX"

	set cmd "exec bin/mf2t T.VSQ | "
	append cmd "bin/Perl \"$Vsqtxt2lipsyncScript\" "
	if {$outputTimeInSec == 0} { append cmd "-i " }
	append cmd "-t \"$tracks\" "
	if {$beatOption != ""} { append cmd "$beatOption " }
	append cmd "-m \"$outputModeOpt\" -p T.TBX \"$fps\" \"$offset\" "
	append cmd "> T.TXT"

	if { [catch {eval $cmd} msg] } {
	    tk_messageBox -icon info -message "conversion error:\n$msg\n" -type ok -parent $w
	    return
	}
	
	if { $outputMode == "m" } {
	    set cmd "exec bin/perl \"$Vocalips2musicxmlScript\" "
	    append cmd "-b $bpm -u 480 -n $numer -d $denom -p $pshift -q $qnt -r $gap < T.TXT > T.XML"
	    if { [catch {eval $cmd} msg] } {
		tk_messageBox -icon info -message "conversion error:\n$msg\n" -type ok -parent $w
		return
	    }
	    file delete "T.TXT"
	    file copy -force "T.XML" "T.TXT"
# hmm... the code below should encode the xml file in utf-8, but it doesnot:
# For instance, Japanese "DO" is converted as follows:
# nkf -w  yields: \343\201\251   , which IS accepted by Sinsy, 
# this    yields: \302\202\303\207,  which IS NOT accepted.
# Maybe more options necessary.
#	    set fin [open "T.XML" r]
#	    set fout [open "T.TXT" w]
#	    fconfigure $fin -encoding cp932
#	    fconfigure $fout -encoding utf-8
#	    puts $fout [read $fin]
#	    close $fin
#	    close $fout

	}

	file copy -force "T.TXT" "$outfile"
	tk_messageBox -icon info -message "Done! $outfile created.\n" -type ok -parent $w
    } else {
	tk_messageBox -icon info -message "save canceled.\n" -type ok -parent $w
    }
}


proc cleanWorkFiles {} {
    catch {file delete "T.MFT"}
    catch {file delete "T1.TXT"}
    catch {file delete "T2.TXT"}
    catch {file delete "T.VSQ"}
    catch {file delete "T.TBX"}
    catch {file delete "T.TXT"}
    catch {file delete "T.XML"}
}	
